/*
 *
 *   Copyright (C) 2005 by Raymond Huang
 *   plushuang at users.sourceforge.net
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  ---
 *
 *  In addition, as a special exception, the copyright holders give
 *  permission to link the code of portions of this program with the
 *  OpenSSL library under certain conditions as described in each
 *  individual source file, and distribute linked combinations
 *  including the two.
 *  You must obey the GNU Lesser General Public License in all respects
 *  for all of the code used other than OpenSSL.  If you modify
 *  file(s) with this exception, you may extend this exception to your
 *  version of the file(s), but you are not obligated to do so.  If you
 *  do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source
 *  files in the program, then also delete it here.
 *
 */

#include <glib.h>
#include <string.h>
#include <urlglib/url_info.h>

// UrlInfo -----------------------------------------------------------

void url_info_init (UrlInfo* uli)
{
	uli->protocol     = NULL;
	uli->id           = NULL;
	uli->password     = NULL;
	uli->address      = NULL;
	uli->port         = NULL;
	uli->path         = NULL;
	uli->filename     = NULL;
	uli->ext_filename = NULL;
	uli->query        = NULL;

	uli->protocol_len = 0;
	uli->id_len       = 0;
	uli->password_len = 0;
	uli->address_len  = 0;
	uli->port_len     = 0;
	uli->path_len     = 0;
	uli->dir_len      = 0;
	uli->filename_len = 0;
	uli->main_filename_len = 0;
	uli->ext_filename_len  = 0;
	uli->query_len    = 0;
}

UrlInfo* url_info_new (const char* url_str)
{
	UrlInfo* uli = g_malloc (sizeof (UrlInfo));

	url_info_part (uli, url_str);
	return uli;
}

UrlInfo* url_info_new_len (const char* url_str, int len)
{
	UrlInfo* uli = g_malloc (sizeof (UrlInfo));

	url_info_part_len (uli, url_str, len);
	return uli;
}

UrlInfo* url_info_new_path (const char* path_str)
{
	UrlInfo* uli = g_malloc (sizeof (UrlInfo));

	url_info_part_path (uli, path_str);
	return uli;
}

UrlInfo* url_info_new_path_len (const char* path_str, int len)
{
	UrlInfo* uli = g_malloc (sizeof (UrlInfo));

	url_info_part_path_len (uli, path_str, len);
	return uli;
}

int  url_info_part (UrlInfo* uli, const char* url_str)
{
	return url_info_part_len (uli, url_str, -1);
}

void url_info_part_path (UrlInfo* uli, const char* path_str)
{
	url_info_part_path_url (uli, path_str, -1, FALSE);
}

int  url_info_part_len (UrlInfo* uli, const char* url_str, int len)
{
	const char* p_cur;
	const char* p_end;
	const char* temp;
	char  character;
	int   continue_flag;

	// initialize
	url_info_init (uli);

	// check parameter
	if (url_str == NULL)
		return FALSE;
	if (len == -1)
		len = strlen (url_str);
	if (len && url_str[0]=='.') {
		url_info_part_path_url (uli, url_str, len, TRUE);
		return FALSE;
	}
	p_end = url_str + len;

	// protocol
	for (p_cur = url_str; p_cur != p_end; ++p_cur) {
		if ( *p_cur == ':' ) {
			if (p_cur != url_str && p_end - p_cur >= 2 &&
			    p_cur[1] == '/' && p_cur[2] == '/' )
			{
				uli->protocol = url_str;
				uli->protocol_len = p_cur - url_str;
				p_cur += 3;
				break;
			}
			return FALSE;
		}
		if ( (*p_cur>='a' && *p_cur<='z') || (*p_cur>='A' && *p_cur<='Z') )
			continue;
	}

	// id:password and address:port
	continue_flag = TRUE;
	for (temp = p_cur; continue_flag; ++temp) {

		character = (temp == p_end) ? '/' : *temp;
		switch (character) {
		case '@':
		case '/':
			continue_flag = FALSE;    // break loop
			if (uli->address) {
				uli->port     = p_cur;
				uli->port_len = temp - p_cur;
				p_cur = temp + 1;
				break;
			}
		case ':':
			uli->address     = p_cur;
			uli->address_len = temp - p_cur;
			p_cur = temp + 1;
			break;
		}
		if (character == '@' && uli->id == NULL) {
    		// swap address:port and id:password
			continue_flag = TRUE;
			uli->id     = uli->address;
			uli->id_len = uli->address_len;
			uli->password     = uli->port;
			uli->password_len = uli->port_len;

			uli->address     = NULL;
			uli->address_len = 0;
			uli->port     = NULL;
			uli->port_len = 0;
		}
		if (temp == p_end)
			return (uli->protocol) ? TRUE : FALSE;
	}

	url_info_part_path_url (uli, p_cur, len - (p_cur-url_str), TRUE);

	return (uli->protocol) ? TRUE : FALSE;
}

void url_info_part_path_url (UrlInfo* uli, const char* path_str, int len, int is_url)
{
	const char* p_cur = path_str;
	const char* p_end;
	const char* temp;
	char  character;
	char  dir_separator;
	int   continue_flag;

	// initialize
	if ( is_url )
		dir_separator = '/';
	else {
		dir_separator = G_DIR_SEPARATOR;
		url_info_init (uli);
	}
	// check parameter
	if (path_str == NULL)
		return;
	if (len == -1)
		len = strlen (path_str);
	p_end = path_str + len;
	uli->path = path_str;

	// dir
	for (temp = p_end-1; temp >= path_str; --temp) {
		if (*temp == dir_separator) {
			uli->dir_len  = temp - path_str + 1;
			uli->path_len = uli->dir_len;
			p_cur = temp + 1;
			break;
		}
	}
	if (p_cur == p_end)
		return;

	// filename and extension name
	uli->filename = p_cur;
	continue_flag = TRUE;
	for (temp = p_cur; continue_flag; ++temp) {

		character = (temp == p_end) ? '?' : *temp;
		switch (character) {
		case '?':
			continue_flag = FALSE;    // break loop
			uli->filename_len  = temp - uli->filename;
			uli->path_len     += uli->filename_len;
			if (uli->main_filename_len) {
				uli->ext_filename     = p_cur;
				uli->ext_filename_len = temp - p_cur;
				p_cur = temp + 1;
				break;
			}
		case '.':
			p_cur = temp + 1;
			uli->main_filename_len = temp - uli->filename;
			break;
		}
		if (temp == p_end)
			return;
	}

	// query
	uli->query = p_cur;
	uli->query_len = p_end - p_cur;
}

static int hex_char_2_int (char chr)
{
	if (chr >= '0' && chr <= '9')
		return chr - '0';
	if (chr >= 'A' && chr <= 'F');
		return chr - 'A' + 10;
//	if (chr >= 'a' && chr <= 'f');
		return chr - 'a' + 10;
}

char* url_info_get_filename (UrlInfo* uli)
{
 	GString* gstr;
	const char* p_cur;
	const char* p_end;
	char  chr;

	if (uli->filename_len == 0)
		return NULL;

	gstr = g_string_sized_new (uli->filename_len);
	p_end = uli->filename + uli->filename_len;

	for (p_cur=uli->filename; p_cur != p_end; p_cur++) {
		chr = *p_cur;
		if (chr == '%' && p_cur+2 < p_end) {
			chr  = hex_char_2_int (p_cur[1]) << 4;
			chr += hex_char_2_int (p_cur[2]);
			p_cur += 2;
		}
		g_string_append_c (gstr, chr);
	}
	return g_string_free (gstr, FALSE);
}

